"""{{ project_name }} URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
import os
from django.contrib import admin
from django.urls import path, include
from django.conf.urls.static import static
from wq.db import rest
from django.conf import settings


urlpatterns = [
    path('admin/', admin.site.urls),
    path('', include(rest.router.urls)),
]

if settings.DEBUG_WITH_RUNSERVER:

    # To use django-media-thumbnailer
    # urlpatterns.append(url('^media/', include('dmt.urls')))

    urlpatterns += static('/media/', document_root=settings.MEDIA_ROOT)

    # after building...
    urlpatterns += static(
        '/', document_root=os.path.join(settings.BASE_DIR, 'htdocs')
    )



"""
Django settings for {{ project_name }} project.

Based on the Django 2.0 template, with wq-specific modifications noted as such.
Generated by 'wq start' {{ wq_start_version }}.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.0/ref/settings/

For more information about wq.db's Django settings see
http://wq.io/docs/settings

"""

import os
from os.path import dirname

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
# wq: extra dirname()s to account for db/ and settings/ folders
BASE_DIR = dirname(dirname(dirname(dirname(os.path.abspath(__file__)))))

# wq: SECRET_KEY, DEBUG, and ALLOWED_HOSTS are defined in dev.py/prod.py


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    {% if not with_gis %}# {% endif %}'django.contrib.gis',
    'rest_framework',

    'wq.db.rest',
    'wq.db.rest.auth',

    # Project apps
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = '{{ project_name }}.urls'

# wq: Recommended settings for Django and rest_framework
from wq.db.default_settings import (
    TEMPLATES,
    SESSION_COOKIE_HTTPONLY,
    REST_FRAMEWORK,
)
TEMPLATES[0]['DIRS'] = [os.path.join(BASE_DIR, 'templates')]

# wq: Recommended settings unique to wq.db
from wq.db.default_settings import (
    ANONYMOUS_PERMISSIONS,
    SRID,
)

WSGI_APPLICATION = '{{ project_name }}.wsgi.application'


# wq: DATABASES is defined in dev.py/prod.py

# Password validation
# https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.0/howto/static-files/

STATIC_URL = '/static/'

# wq: Configure paths for default project layout
STATIC_ROOT = os.path.join(BASE_DIR, 'htdocs', 'static')
MEDIA_ROOT = os.path.join(BASE_DIR, 'media'){% if with_npm %}
WQ_SCRIPT_FILE = os.path.join(BASE_DIR, 'htdocs', 'index.html'){% endif %}
VERSION_TXT = os.path.join(BASE_DIR, 'version.txt')
MEDIA_URL = '/media/'



from wq.core import wq
import click
import os
import shutil
import subprocess
from django.core.management import call_command
from django.core.management.commands import startproject
from pkg_resources import resource_filename
from wq.core.info import print_versions
from .version import VERSION


template = resource_filename('wq.start', 'django_project')
# resource_filename not returning absolute path after pip install
if os.sep not in template:
    import wq as wq_module
    template = wq_module.__path__[0] + os.sep + template


class StartProjectCommand(startproject.Command):
    def add_arguments(self, parser):
        super(StartProjectCommand, self).add_arguments(parser)
        parser.add_argument('--domain', help="Web Domain")
        parser.add_argument('--app-id', help="App Identifier")
        parser.add_argument('--with-gis', help="Enable GeoDjango")
        parser.add_argument('--with-npm', help="Enable NPM")
        parser.add_argument('--with-pgb', help="Enable PhoneGap Build")
        parser.add_argument('--wq-start-version', help="wq start version")


@wq.command()
@click.argument("project_name", required=False)
@click.argument("destination", required=False)
@click.option(
    "-d", "--domain", help='Web domain (e.g. example.wq.io)'
)
@click.option(
    "--with-gis/--without-gis", default=None, help="Enable GeoDjango"
)
@click.option(
    "--with-npm/--without-npm", default=None,
    help="Enable NPM (& Create React App)"
)
@click.option(
    "--npm-install/--skip-npm-install",  default=None,
    help="Run NPM install after creating project"
)
@click.option(
    '--with-pgb/--without-pgb', default=None, help="Enable PhoneGap Build",
)
@click.option(
    "-i", "--app-id", help="Application ID (e.g. io.wq.example)"
)
def start(project_name, destination, domain=None,
          with_gis=None, with_npm=None, npm_install=None,
          with_pgb=None, app_id=None):
    """
    Start a new project with wq.app and wq.db.  A new Django project will be
    created from a wq-specific template.  Any options not specified via
    arguments will be prompted for instead.

    After running this command, you may want to do the following:

    \b
        sudo chown www-data media/
        ./deploy.sh 0.0.0

    See https://wq.io/docs/setup for more tips on getting started with wq.
    """
    any_prompts = False

    if project_name is None:
        any_prompts = True
        project_name = click.prompt(
            'Project codename',
        )

    if destination is None:
        any_prompts = True
        destination = click.prompt(
            'Directory',
            default='./{}/'.format(project_name),
        )

    if domain is None:
        any_prompts = True
        domain = click.prompt(
            'Web domain',
            default='{}.example.org'.format(project_name)
        )

    if with_gis is None:
        any_prompts = True
        with_gis = click.confirm(
            'Enable GIS? (Requires PostGIS or SpatialLite)',
            default=False,
        )

    if with_npm is None:
        any_prompts = True
        with_npm = click.confirm(
            'Enable NPM / Create React App? (Requires Node.js)',
            default=False,
        )

    if app_id:
        with_pgb = True

    if with_pgb is None:
        any_prompts = True
        with_pgb = click.confirm(
            'Enable PhoneGap Build? (Requires Adobe account)',
            default=False,
        )

    if with_pgb and app_id is None:
        any_prompts = True
        app_id = click.prompt(
            'Application Bundle ID',
            default='.'.join(reversed(domain.split('.')))
        )

    os.makedirs(
        os.path.abspath(os.path.expanduser(destination)),
        exist_ok=True,
    )
    args = [project_name, destination]
    kwargs = dict(
        template=template,
        extensions="py,yml,conf,html,sh,js,css,json,xml,gitignore".split(","),
        domain=domain,
        wq_start_version=VERSION,
        with_gis=with_gis,
        with_npm=with_npm,
        with_pgb=with_pgb,
        app_id=app_id,
    )
    call_command(StartProjectCommand(), *args, **kwargs)

    path = destination or project_name
    print_versions(os.path.join(path, 'requirements.txt'))
    shutil.copytree(
        resource_filename('xlsconv', 'templates'),
        os.path.join(path, 'master_templates'),
        ignore=shutil.ignore_patterns("*.py-tpl"),
    )

    if not with_pgb:
        shutil.rmtree(os.path.join(path, 'app', 'pgb'))

    if with_npm:
        shutil.rmtree(os.path.join(path, 'app', 'js'))
        os.remove(os.path.join(path, 'app', 'src', 'README.md'))
    else:
        shutil.rmtree(os.path.join(path, 'app', 'src'))
        os.remove(os.path.join(path, 'app', 'js', 'README.md'))
        os.remove(os.path.join(path, 'app', 'package.json'))
        os.remove(os.path.join(path, 'app', 'public', 'index.html'))
        if with_pgb:
            os.remove(os.path.join(path, 'app', 'pgb', 'pginit.js'))
        public_dir = os.path.join(path, 'app', 'public')
        for filename in os.listdir(public_dir):
            os.rename(
                os.path.join(public_dir, filename),
                os.path.join(path, 'app', filename),
            )
        os.rmdir(public_dir)

    flags = []
    if with_gis:
        flags.append('GIS')
    if with_npm:
        flags.append('NPM')
    if with_pgb:
        flags.append('PGB')

    if len(flags) == 3:
        flag_summary = " with {0}, {1}, and {2} support".format(*flags)
    elif len(flags) == 2:
        flag_summary = " with {0} and {1} support".format(*flags)
    elif len(flags) == 1:
        flag_summary = " with {} support".format(*flags)
    else:
        flag_summary = ""

    if any_prompts:
        click.echo()
    click.echo(
        'Project "{project_name}" created successfully in {destination}'
        "{flag_summary}.".format(
            project_name=project_name,
            destination=destination,
            flag_summary=flag_summary,
        )
    )
    if with_npm:
        if npm_install is None:
            npm_install = click.confirm(
                "Run npm install now?"
            )
        if npm_install:
            subprocess.run(
                ['npm', 'install'],
                cwd=os.path.join(path, 'app')
            )
            click.echo()
            click.echo("npm install complete.")

    if with_npm and not npm_install:
        click.echo(
            "After running npm install in ./app/,"
            " run ./deploy.sh 0.0.0 to finish initial setup."
        )
    else:
        click.echo("Run ./deploy.sh 0.0.0 to finish initial setup.")



from .base import *


# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '{{ secret_key }}'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

# wq: Determine if we are running off django's testing server
DEBUG_WITH_RUNSERVER = False

ALLOWED_HOSTS = ["{{ domain }}"]


# Database
# https://docs.djangoproject.com/en/2.0/ref/settings/#databases

DATABASES = {
    'default': {
        {% if with_gis %}'ENGINE': 'django.contrib.gis.db.backends.postgis',
        {% else %}'ENGINE': 'django.db.backends.postgres',
        # To enable GeoDjango:
        # 'ENGINE': 'django.contrib.gis.db.backends.postgis',
        {% endif %}'NAME': '{{ project_name }}',
        'USER': '{{ project_name }}',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}



from wq.core import wq
import click
import os
from xlsconv import parse_xls, xls2django, xls2html, html_context, render
from pkg_resources import resource_filename
import subprocess
import json
from difflib import unified_diff
from pyxform.question_type_dictionary import QUESTION_TYPE_DICT as QTYPES


templates = resource_filename('wq.start', 'master_templates')
# resource_filename not returning absolute path after pip install
if os.sep not in templates:
    import wq as wq_module
    templates = wq_module.__path__[0] + os.sep + templates


@wq.command()
@click.argument(
    "xlsform",
    required=True,
    type=click.Path(exists=True),
)
@click.option(
    "--input-dir",
    default="../master_templates",
    type=click.Path(exists=True),
    help="Source / master templates",
)
@click.option(
    "--django-dir",
    default=".",
    type=click.Path(exists=True),
    help="Root of Django project",
)
@click.option(
    "--template-dir",
    default="../templates",
    type=click.Path(exists=True),
    help="Path to shared template directory",
)
@click.option(
    "--form-name",
    help="Name to use for Django app and template prefix",
)
@click.option(
    "--with-admin/--no-admin",
    default=False,
    help="Generate admin.py",
)
@click.option(
    "--force", "-f",
    is_flag=True,
    default=False,
    help="Answer yes to all prompts",
)
def addform(xlsform, input_dir, django_dir, template_dir,
            form_name, with_admin, force):
    """
    Convert an XLSForm into a Django app for wq.  Generates Python and mustache
    files including:

    \b
        db/[form_name]/models.py
        db/[form_name]/rest.py
        templates/[form_name]_detail.html
        templates/[form_name]_edit.html
        templates/[form_name]_list.html
    """

    xls_json = parse_xls(xlsform)
    if not form_name:
        form_name = xls_json['name']

    if not os.path.exists(os.path.join(django_dir, form_name)):
        os.mkdir(os.path.join(django_dir, form_name))

    if not os.path.exists(os.path.join(django_dir, form_name, 'migrations')):
        os.mkdir(os.path.join(django_dir, form_name, 'migrations'))

    create_file(
        [django_dir, form_name, '__init__.py'],
        "",
        overwrite=force,
    )

    create_file(
        [django_dir, form_name, 'migrations', '__init__.py'],
        "",
        overwrite=force,
    )

    create_file(
        [django_dir, form_name, 'models.py'],
        xls2django(xlsform, 'models'),
        overwrite=force,
    )

    has_nested = False
    for field in xls_json['children']:
        if field.get('wq:nested', False):
            has_nested = True
    if has_nested:
        create_file(
            [django_dir, form_name, 'serializers.py'],
            xls2django(xlsform, 'serializers'),
            overwrite=force,
        )

    create_file(
        [django_dir, form_name, 'rest.py'],
        xls2django(xlsform, 'rest'),
        overwrite=force,
    )
    if with_admin:
        create_file(
            [django_dir, form_name, 'admin.py'],
            xls2django(xlsform, 'admin'),
            overwrite=force,
        )

    template_types = set(['detail', 'edit', 'list'])
    for field in xls_json['children']:
        if 'geo' in field['type']:
            if 'popup' in template_types:
                print("Warning: multiple geometry fields found.")
            template_types.add('popup')
    for tmpl in template_types:
        create_file(
            [template_dir, "%s_%s.html" % (form_name, tmpl)],
            xls2html(xlsform, os.path.join(input_dir, "%s.html" % tmpl)),
            overwrite=force,
        )

    settings_path = None
    for path, dirs, files in os.walk(django_dir):
        for filename in files:
            if filename == 'settings.py':
                settings_path = (path, filename)
            elif path.endswith('/settings') and filename == 'base.py':
                settings_path = (path[:-9], 'settings', filename)

    if not settings_path:
        return

    new_settings = []
    app_section = False
    has_app = False
    for row in open(os.path.join(*settings_path)):
        if 'INSTALLED_APPS' in row:
            app_section = True
        elif app_section:
            if ')' in row or ']' in row:
                app_section = False
                if not has_app:
                    new_settings.append(
                        "    '%s',%s" % (form_name, os.linesep)
                    )
            else:
                if '"%s"' % form_name in row or "'%s'" % form_name in row:
                    has_app = True
        new_settings.append(row)
    create_file(
        settings_path, "".join(new_settings), overwrite=force, show_diff=True
    )
    result = subprocess.check_output(
        [os.path.join(django_dir, 'manage.py'), 'makemigrations']
    ).decode('utf-8').strip()
    print(result)
    if 'No changes' in result:
        return
    migrate = force or click.confirm("Update database schema?", default=True)
    if not migrate:
        return
    subprocess.call(
        [os.path.join(django_dir, 'manage.py'), 'migrate']
    )


@wq.command()
@click.option(
    "--input-dir",
    default="../master_templates",
    type=click.Path(exists=True),
    help="Source / master templates",
)
@click.option(
    "--django-dir",
    default=".",
    type=click.Path(exists=True),
    help="Root of Django project",
)
@click.option(
    "--template-dir",
    default="../templates",
    type=click.Path(exists=True),
    help="Path to shared template directory",
)
@click.option(
    '-f',
    '-overwrite',
    default=False,
    is_flag=True,
    help="Replace existing templates",
)
def maketemplates(input_dir, django_dir, template_dir, overwrite):
    """
    Generate mustache templates for wq.db.rest.  Automatically discovers all
    registered models through ./manage.py dump_config.

    \b
        templates/[model_name]_detail.html
        templates/[model_name]_edit.html
        templates/[model_name]_list.html
    """
    result = subprocess.check_output(
        [os.path.join(django_dir, 'manage.py'), 'dump_config']
    )
    config = json.loads(result.decode('utf-8'))

    has_diff = False

    for page in config['pages'].values():
        if not page.get('list', None):
            continue

        if has_diff == 'skipall':
            break

        def process_fields(fields):
            for field in fields:
                if field['type'] in ('repeat', 'group'):
                    field['wq:nested'] = True
                    if field['type'] == 'repeat':
                        field['wq:many'] = True
                    process_fields(field['children'])
                elif field['type'] in QTYPES:
                    field['type_info'] = QTYPES[field['type']]
                else:
                    field['type_info'] = {'bind': {'type': field['type']}}

        process_fields(page['form'])
        context = html_context({
            'name': page['name'],
            'title': page['name'],
            'children': page['form'],
        })
        context['form']['urlpath'] = page['url']

        template_types = set(page.get('modes', ['detail', 'edit', 'list']))

        has_geo = False
        for field in page['form']:
            if 'geo' in field['type']:
                if has_geo:
                    print("Warning: multiple geometry fields found.")
                has_geo = True

        if page.get('map', None):
            has_geo = True

        if has_geo:
            context['form']['has_geo'] = True
            template_types.add('popup')

        for tmpl in template_types:
            has_diff = create_file(
                [template_dir, "%s_%s.html" % (page['name'], tmpl)],
                render(context, os.path.join(input_dir, '%s.html' % tmpl)),
                overwrite=overwrite,
                previous_diff=has_diff,
            )
            if has_diff == 'skipall':
                break


def create_file(path, contents, overwrite=False,
                show_diff=False, previous_diff=False):
    filename = os.path.join(*path)
    has_diff = previous_diff
    if os.path.exists(filename) and not overwrite:
        existing_file = open(filename, 'r')
        existing_content = existing_file.read()
        if existing_content.strip() == contents.strip():
            return False

        def print_diff():
            diff = unified_diff(
                existing_content.split('\n'),
                contents.split('\n'),
                fromfile="%s (current)" % path[-1],
                tofile="%s (new)" % path[-1],
            )
            for row in diff:
                print(row)

        if show_diff:
            print_diff()
            message = "Update %s? [Y/n/d/?]"
            default_choice = 'y'
        else:
            if not previous_diff:
                choice = click.prompt('Update templates? [y/n]')
                if choice.lower() != 'y':
                    print("Skipping template updates.")
                    return 'skipall'

            message = "%s already exists; overwrite? [y/n/d/?]"
            default_choice = None

        has_diff = True
        choice = ''
        while choice.lower() not in ('y', 'n'):
            if path[-2] == 'settings':
                filename = os.path.join(*path[-2:])
            else:
                filename = path[-1]
            choice = click.prompt(
                message % filename, default=default_choice, show_default=False
            )
            if choice == '' and show_diff:
                choice = 'y'
            if choice.lower() == 'n':
                return has_diff
            elif choice.lower() == '?':
                print(
                    '  y - overwrite\n'
                    '  n - skip\n'
                    '  d - show diff\n'
                    '  ? - show help'
                )
            elif choice.lower() == 'd':
                print_diff()
    out = open(os.path.join(*path), 'w')
    out.write(contents)
    out.close()
    return has_diff



#!/usr/bin/env python
import sys
import json

file1, file2 = sys.argv[1:]

json1 = json.load(open(file1))
json2 = json.load(open(file2))

if 'debug' in json2:
    json2.pop('debug')

def unpack_diff(json1, json2, prefix=''):
    if isinstance(json1, list) and isinstance(json2, list):
        for i, (val1, val2) in enumerate(zip(json1, json2)):
            unpack_diff(val1, val2, prefix + '[%s]' % i)
        if len(json1) != len(json2):
            print('len(%s)' % prefix, len(json1), '!=', len(json2))

    elif isinstance(json1, dict) and isinstance(json2, dict):
        if prefix:
            prefix += '.'
        for key in sorted(set(json1.keys()) | set(json2.keys())):
            val1 = json1.get(key)
            val2 = json2.get(key)
            unpack_diff(val1, val2, prefix + key)

    elif json1 != json2:
        print(prefix, json1, '!=', json2)

if json1 != json2:
    unpack_diff(json1, json2)
    sys.exit("Files do not match!")



import os
import sys
from .base import *


# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '{{ secret_key }}'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

# wq: Determine if we are running off django's testing server
DEBUG_WITH_RUNSERVER = 'manage.py' in sys.argv[0]

{% if with_npm %}
if DEBUG_WITH_RUNSERVER:
    STATICFILES_DIRS = [
        os.path.join(BASE_DIR, 'app', 'build', 'static')
    ]
{% endif %}

ALLOWED_HOSTS = []

# Database
# https://docs.djangoproject.com/en/2.0/ref/settings/#databases

DATABASES = {
    'default': {
        {% if with_gis %}'ENGINE': 'django.contrib.gis.db.backends.spatialite',
        {% else %}'ENGINE': 'django.db.backends.sqlite3',
        # To enable GeoDjango:
        # 'ENGINE': 'django.contrib.gis.db.backends.spatialite',
        {% endif %}'NAME': os.path.join(BASE_DIR, 'conf', '{{ project_name }}.sqlite3'),
    }
}

{% if not with_gis %}# {% endif %}SPATIALITE_LIBRARY_PATH = 'mod_spatialite.so'



import os
from setuptools import setup

LONG_DESCRIPTION = """
Project scaffolding tools for creating a new crowdsourcing or citizen science application with the wq framework.
"""


def readme():
    try:
        readme = open('README.md')
    except IOError:
        return LONG_DESCRIPTION
    return readme.read()


def create_wq_namespace():
    """
    Generate the wq namespace package
    (not checked in, as it technically is the parent of this folder)
    """
    if os.path.isdir("wq"):
        return
    os.makedirs("wq")
    init = open(os.path.join("wq", "__init__.py"), 'w')
    init.write("__import__('pkg_resources').declare_namespace(__name__)")


def list_package_data(root):
    """
    Include project template as package data
    """
    paths = []

    for base, dirs, files in os.walk(root, topdown=True):
        if 'node_modules' in dirs:
            dirs.remove('node_modules')
        paths.extend([
            os.path.join(base, name) for name in files
            if name not in ('.git', 'package-lock.json')
        ])
    return paths

create_wq_namespace()

# Project template data
TEMPLATES = [
    'django_project',
]
TEMPLATE_DATA = []
for folder in TEMPLATES:
    TEMPLATE_DATA.extend(list_package_data(folder))

setup(
    name='wq.start',
    use_scm_version=True,
    author='S. Andrew Sheppard',
    author_email='andrew@wq.io',
    url='https://wq.io/wq.start',
    license='MIT',
    description=LONG_DESCRIPTION.strip(),
    long_description=readme(),
    long_description_content_type='text/markdown',
    entry_points={'wq': 'wq.start=wq.start'},
    packages=['wq.start'],
    package_dir={'wq.start': '.'},
    namespace_packages=['wq'],
    package_data={'wq.start': TEMPLATE_DATA},
    install_requires=[
        'wq.core',
        'psycopg2-binary',
        'xlsconv>=1.2.0',
    ],
    setup_requires=[
        'setuptools_scm',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: JavaScript',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Pre-processors',
    ]
)
